/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "config.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "dbacl.h"


/* most functions in this file are logically identical in wide char
 * and multibyte versions, except that char is replaced by wchar_t
 * etc. It's become unwieldy to keep two slightly different copies of
 * all functions involved, so we use the preprocessor to build a poor
 * man's template facility.
 * 
 * The "template" macros work as follows:
 * mbw_lit("abc")             -> "abc" or L"abc"
 * mbw_t                      -> char  or wchar_t
 * mbw_prefix(good_char)(x)   -> good_char(x) or w_good_char(x)
 *
 * Once the template macros have done their work, we obtain ordinary
 * functions named in a parallel fashion, where the wide character
 * functions have a w_ prefix, and instances of char are substituted
 * with instances of wchar_t.
 *
 * The code below is split into uncommon code, where the
 * implementations of corresponding functions is different, and common
 * code where the implementation is identical. Only identical code is
 * "templatized".
 */

#if defined MBW_WIDE

#define mbw_lit(x) L##x
#define mbw_t wchar_t
#define mbw_prefix(f) w_##f
#define mbw_isalnum(c) iswalnum(c)
#define mbw_isalpha(c) iswalpha(c)
#define mbw_isgraph(c) iswgraph(c)
#define mbw_isspace(c) iswspace(c)
#define mbw_iscntrl(c) iswcntrl(c)
#define mbw_ispunct(c) iswpunct(c)
#define mbw_isxdigit(c) iswxdigit(c)
#define mbw_isdigit(c) iswdigit(c)
#define mbw_strncmp(x,y,z) wcsncmp(x,y,z)
#define mbw_strncasecmp(x,y,z) wcsncasecmp(x,y,z)
#define mbw_strtol(x,y,z) wcstol(x,y,z)
#define mbw_tolower(x) towlower(x)
#define mbw_strchr(x,y) wcschr(x,y)
#define mbw_strcpy(x,y) wcscpy(x,y)
#define mbw_strlen(x) wcslen(x)
#define mbw_isprint(x) iswprint(x)

#elif defined MBW_MB

#define mbw_lit(x) x
#define mbw_t char
#define mbw_prefix(f) f
#define mbw_isalnum(c) isalnum((unsigned char)c)
#define mbw_isalpha(c) isalpha((unsigned char)c)
#define mbw_isgraph(c) isgraph((unsigned char)c)
#define mbw_isspace(c) isspace((unsigned char)c)
#define mbw_iscntrl(c) iscntrl((unsigned char)c)
#define mbw_ispunct(c) ispunct((unsigned char)c)
#define mbw_isxdigit(c) isxdigit((unsigned char)c)
#define mbw_isdigit(c) isdigit((unsigned char)c)
#define mbw_strncmp(x,y,z) strncmp(x,y,z)
#define mbw_strncasecmp(x,y,z) strncasecmp(x,y,z)
#define mbw_strtol(x,y,z) strtol(x,y,z)
#define mbw_tolower(x) tolower(x)
#define mbw_strchr(x,y) strchr(x,y)
#define mbw_strcpy(x,y) strcpy(x,y)
#define mbw_strlen(x) strlen(x)
#define mbw_isprint(x) isprint(x)

#else

#error "This file must be compiled with -DMBW_WIDE or -DMBW_P"

#endif

extern options_t u_options;
extern options_t m_options;
extern MBOX_State mbox;
extern XML_State xml;

typedef struct {
  mbw_t *cache;
  mbw_t *data_ptr;
  size_t cache_len;
  size_t max_line_len;
} mbw_prefix(decoding_cache);

mbw_prefix(decoding_cache) mbw_prefix(b64_dc) = {NULL, NULL, 0, 0};
mbw_prefix(decoding_cache) mbw_prefix(qp_dc) = {NULL, NULL, 0, 0};

/* uncommon code */

/***********************************************************
 * UTILITY FUNCTIONS                                       *
 ***********************************************************/

#if defined HAVE_MBRTOWC && defined MBW_WIDE
/* compiler doesn't seem to know this function is in the 
 * library, so we define our own - bug or just plain weird? */
int wcsncasecmp(const wchar_t *s1, const wchar_t *s2, size_t n) {
  register size_t i = 0;
  while( i < n ) {
    if( tolower(*s1) != tolower(*s2) ) {
      return towlower(*s1) - towlower(*s2);
    }
    s1++;
    s2++;
  }
  return 0;
}
#endif


#if defined HAVE_MBRTOWC && defined MBW_WIDE

int w_b64_code(wchar_t c) {
  if( (c >= L'A') && (c <= L'Z') ) {
    return (c - L'A');
  } else if( (c >= L'a') && (c <= L'z') ) {
    return (c - L'a') + 26;
  } else if( (c >= L'0') && (c <= L'9') ) {
    return (c - L'0') + 52;
  } else if( c == L'+' ) { 
    return 62;
  } else if( c == L'/' ) {
    return 63;
  } else if( c == L'=' ) {
    return 64;
  } else {
    return -1;
  }
}

int w_qp_code(wchar_t c) {
  if( (c >= L'0') && (c <= L'9') ) {
    return (c - L'0');
  } else if( (c >= L'A') && (c <= L'F') ) {
    return (c - L'A') + 10;
  } else {
    return -1;
  }
}

#else 

/* warning: only use char here so we never have to bother about endinanness */
static const signed char b64_code_table[256] = { 
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,62,-1,-1,-1,63,52,53,
  54,55,56,57,58,59,60,61,-1,-1,
  -1,64,-1,-1,-1, 0, 1, 2, 3, 4,
   5, 6, 7, 8, 9,10,11,12,13,14,
  15,16,17,18,19,20,21,22,23,24,
  25,-1,-1,-1,-1,-1,-1,26,27,28,
  29,30,31,32,33,34,35,36,37,38,
  39,40,41,42,43,44,45,46,47,48,
  49,50,51,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1
};

#define b64_code(c) ((int)b64_code_table[(int)c])

static const signed char qp_code_table[256] = {
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1, 0, 1,
   2, 3, 4, 5, 6, 7, 8, 9,-1,-1,
  -1,-1,-1,-1,-1,10,11,12,13,14,
  15,16,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1
};

#define qp_code(c) (qp_code_table[(int)c])

#endif


/* common code */

#if defined MBW_MB || (defined MBW_WIDE && defined HAVE_MBRTOWC)

/***********************************************************
 * UTILITY FUNCTIONS                                       *
 ***********************************************************/

/* 
 * this code generates mystrcasestr() and w_mystrcasestr() 
 * (similar to strstr, but case insensitive)
 */
const mbw_t *mbw_prefix(mystrcasestr)(const mbw_t *haystack, const mbw_t *needle) {
  const mbw_t *p, *q, *r;

  for(p = haystack; *p; p++) {
    q = needle; r = p;
    while( *q && *r && ((tolower(*q) - tolower(*r)) == 0) ) {
      q++; r++;
    }
    if( !*q ) {
      return p;
    }
  }
  return NULL;
}

/***********************************************************
 * DECODING CACHE FUNCTIONS                                *
 ***********************************************************/

void mbw_prefix(init_dc)(mbw_prefix(decoding_cache) *dc, size_t len) {
  if( !dc->cache ) {
    dc->cache = (mbw_t *)malloc(len * sizeof(mbw_t));
    dc->data_ptr = dc->cache;
    dc->cache_len = dc->cache ? len : 0;
    dc->max_line_len = len; 
  }
}

void mbw_prefix(init_decoding_caches)() {
  mbw_prefix(init_dc)(&mbw_prefix(b64_dc),BUFLEN);
  mbw_prefix(init_dc)(&mbw_prefix(qp_dc),BUFLEN);
}

void mbw_prefix(free_decoding_caches)() {
  if( mbw_prefix(b64_dc).cache ) { free(mbw_prefix(b64_dc).cache); }
  if( mbw_prefix(qp_dc).cache ) { free(mbw_prefix(qp_dc).cache); }
}

void mbw_prefix(adjust_cache_size)(mbw_prefix(decoding_cache) *dc, size_t n) {
  mbw_t *p;
  size_t m = (dc->data_ptr - dc->cache);
  while( dc->cache_len < m + n ) {
    p = (mbw_t *)realloc(dc->cache, 2 * dc->cache_len * sizeof(mbw_t));
    if( p ) {
      dc->cache = p;
      dc->data_ptr = p + m;
      dc->cache_len *= 2;
    } else {
      break;
    }
  }
  dc->max_line_len = (dc->max_line_len < n) ? n : dc->max_line_len;
}

bool_t mbw_prefix(flush_cache)(mbw_prefix(decoding_cache) *dc, mbw_t *line, bool_t all) {
  mbw_t *q;
  mbw_t *p;
  if( dc->cache && (dc->data_ptr > dc->cache) ) {

    /* never output more bytes than will fit on output_line */
    p = (dc->data_ptr > dc->cache + dc->max_line_len) ? 
      (dc->cache + dc->max_line_len): dc->data_ptr;

    if( !all ) {
      /* we brake the line at the last space */
      while( !mbw_isspace(*p) && (p > dc->cache) ) {
	p--; 
      }
    }

    if( p > dc->cache ) {
      for(q = dc->cache; q < p; q++) {
	*line++ = *q;
      }
      *line = mbw_lit('\0');

      dc->data_ptr = dc->cache;
      if( !all ) {
	/* now fold unused part back into cache. Note that
	 * b64_line_cache is always NUL terminated, so we don't
	 * need b64_cache_ptr to mark the end. */
	while( *p ) {
	  *dc->data_ptr++ = *p++;
	}
      }
      *dc->data_ptr = mbw_lit('\0');
      return 1;
    }
  }
  return 0;
}

/***********************************************************
 * DECODING FUNCTIONS                                      *
 ***********************************************************/

#define SET1(c) ( (*(c) == mbw_lit('-')) || (*(c) == mbw_lit('.')) )
/* #define SET1(c) ( (*(c) == mbw_lit('-')) || (*(c) == mbw_lit('+')) || (*(c) == mbw_lit('.')) || (*(c) == mbw_lit('_')) || (*(c) == mbw_lit(',')) || (*(c) >= 0xA0) ) */
/* #define SET1(c) ( (*(c) == mbw_lit('-')) || (*(c) == mbw_lit('+')) || (*(c) == mbw_lit('.')) || (*(c) == mbw_lit('_')) || (*(c) == mbw_lit(',')) || (*(c) == mbw_lit('$')) || (*(c) >= 0xA0) ) */
#define SET2(c) ( (*(c) == mbw_lit(',')) || (*(c) == mbw_lit('.')) )
#define SET3(c) ( (*(c) > mbw_lit(' ')) && (*(c) <= mbw_lit('~')) && (*(c) != mbw_lit('>')) )

#define IO(c) ((*(c) & 0xC0) == 0x80)
#define I2O(c) ((*(c) & 0xE0) == 0xC0)
#define I3O(c) ((*(c) & 0xF0) == 0xE0)
#define I4O(c) ((*(c) & 0xF8) == 0xF0)
#define I5O(c) ((*(c) & 0xFC) == 0xF8)
#define I6O(c) ((*(c) & 0xFE) == 0xFC)

#define RANGE(c,x,y) ((*(c) >= x) && (*(c) <= y))
#define DRANGE(c,x,y,u,v) (RANGE(c,x,y) || RANGE(c,u,v))
#define DTEST(s,t,r) (s && t && (cef_state = r))
#define TTEST(s,t,u,r) (s && t && u && (cef_state = r))
#define QTEST(s,t,u,v,r) (s && t && u && v && (cef_state = r))
#define VTEST(s,t,u,v,w,r) (s && t && u && v && w && (cef_state = r))
#define STEST(s,t,u,v,w,x,r) (s && t && u && v && w && x && (cef_state = r))

#define Shift_JIS(c) ( DTEST(DRANGE(c,0x81,0x9F,0xE0,0xFC),DRANGE(c+1,0x40,0x7E,0x80,0xFC),fShift_JIS_1) )

#define EUC_Japanese(c) ( DTEST(RANGE(c,0xA1,0xFE),RANGE(c+1,0xA1,0xFE),fEUC_Japanese_1) || DTEST((*c == 0x8E),RANGE(c+1,0xA0,0xDF),fEUC_Japanese_1) || TTEST((*c == 0x8F),RANGE(c+1,0xA1,0xFE),RANGE(c+2,0xA1,0xFE),fEUC_Japanese_2) )

#define BIG5(c) ( DTEST(RANGE(c,0xA1,0xFE),DRANGE(c+1,0x40,0x7E,0xA1,0xFE),fBIG5_1) )

#define BIG5P(c) ( DTEST(RANGE(c,0x81,0xFE),DRANGE(c+1,0x40,0x7E,0x80,0xFE),fBIG5P_1) )

#define EUC_CN(c) ( DTEST(RANGE(c,0xA1,0xFE),RANGE(c+1,0xA1,0xFE),fEUC_CN_1) )

#define EUC_TW(c) ( DTEST(RANGE(c,0xA1,0xFE),RANGE(c+1,0xA1,0xFE),fEUC_TW_1) || QTEST((*c == 0x8E),RANGE(c+1,0xA1,0xB0),RANGE(c+2,0xA1,0xFE),RANGE(c+3,0xA1,0xFE),fEUC_TW_3) )

#define Johab(c) ( DTEST(RANGE(c,0x84,0xD3),DRANGE(c+1,0x41,0x7E,0x81,0xFE),fJohab_1) || DTEST(DRANGE(c,0xD8,0xDE,0xE0,0xF9),DRANGE(c+1,0x31,0x7E,0x91,0xFE),fJohab_1) )

#define UTF8(c) ( DTEST(I2O(c),IO(c+1),fUTF8_1) || TTEST(I3O(c),IO(c+1),IO(c+2),fUTF8_2) ||  QTEST(I4O(c),IO(c+1),IO(c+2),IO(c+3),fUTF8_3) || VTEST(I5O(c),IO(c+1),IO(c+2),IO(c+3),IO(c+4),fUTF8_4) || STEST(I6O(c),IO(c+1),IO(c+2),IO(c+3),IO(c+4),IO(c+5),fUTF8_5) )

static enum {
  fDEF = 1,
  fANX,
  fNAX,
  fMUL,
  fCUR,
  fADD,
  fSEP_1, fSEP_2, fSEP_3,
  fUTF8_1, fUTF8_2, fUTF8_3, fUTF8_4, fUTF8_5, 
  fShift_JIS_1,
  fEUC_Japanese_1, fEUC_Japanese_2,
  fBIG5_1,
  fBIG5P_1,
  fEUC_CN_1,
  fEUC_TW_1,fEUC_TW_2,fEUC_TW_3,
  fJohab_1
} cef_state = fDEF;

/* some common encoding formats. 
 * 1 = good
 * 0 = ungood
 * 2 = good + end of token
 */
bool_t mbw_prefix(is_cef_char)(register mbw_t *c) {
  switch(cef_state) {
  case fDEF:
    
#if defined MBW_MB
    /* this doesn't make sense for wide characters */
    if(*c & 0x80) {
      if( Shift_JIS(c) || EUC_Japanese(c) || 
	  BIG5(c) || BIG5P(c) || 
	  EUC_CN(c) || EUC_TW(c) ||
	  Johab(c) || UTF8(c) ) { 
 	return 1; 
      } else if( *c < 0xA0 ) { 
	/* not iso-8859-* 
	 * (iso-8859-* characters are filtered by locale functions below)
	 */
	return 0;
      }
    }
#endif

    if( mbw_isalpha(*c) ) {
      if( !(m_options & (1<<M_OPTION_CASEN)) ) {
	*c = mbw_tolower(*c);
      }
      if( mbw_isdigit(c[1]) ) {
	cef_state = fANX;
	return 1;
      } else if( SET1(c+1) && mbw_isalpha(c[2]) ) {
	cef_state = fMUL;
	return 1;
      }
      return 1;
/*     } else if( mbw_isdigit(*c) ) { */
/*       if( mbw_isdigit(c[1]) ) { */
/* 	return 1; */
/*       } else if( mbw_isalpha(c[1]) ) { */
/* 	cef_state = fNAX; */
/* 	return 1; */
/*       }  */
/*       return 2; */
    } else if( mbw_ispunct(*c) ) {
      if( (c[1] == c[0]) && (c[2] == c[0]) ) {
	cef_state = fSEP_3;
	return 1;
      } else {
	switch(*c) {
	case mbw_lit('<'):
	  for(c++; SET3(c); c++);
	  if( *c == mbw_lit('>') ) {
	    cef_state = fADD;
	  }
	  return 0;
	case mbw_lit('$'):
	case mbw_lit(''):
	  if( mbw_isdigit(c[1]) && 
	      (!mbw_isdigit(c[2]) || !mbw_isdigit(c[3]) || !mbw_isdigit(c[4])) ) {
	    cef_state = fCUR;
	    return 1;
	  }
	  return 0;
	default:
	  return 0;
	}
      }
    }
    return 0;
  case fANX:
    if( !mbw_isdigit(c[1]) ) {
      cef_state = fDEF;
      return 2;
    }
    return 1;
  case fNAX:
    if( !(m_options & (1<<M_OPTION_CASEN)) ) {
      *c = mbw_tolower(*c);
    }
    if( !mbw_isalpha(c[1]) ) {
      cef_state = fDEF;
      return 2;
    }
    return 1;
  case fMUL:
    if( !(m_options & (1<<M_OPTION_CASEN)) ) {
      *c = mbw_tolower(*c);
    }
    if( mbw_isalpha(c[1]) ) {
      return 1;
    } else if( SET1(c+1) && mbw_isalpha(c[2]) ) {
      return 1;
    }
    cef_state = fDEF;
    return 2;
  case fADD:
    if( !(m_options & (1<<M_OPTION_CASEN)) ) {
      *c = mbw_tolower(*c);
    }
    switch(*c) {
    case mbw_lit('@'):
/*     case mbw_lit('#'): */
/*     case mbw_lit('?'): */
/*     case mbw_lit('&'): */
/*     case mbw_lit(':'): */
/*     case mbw_lit('/'): */
      return 0;
    case mbw_lit('>'):
      cef_state = fDEF;
      return 0;
    default:
      return 1;
    }
    return 0;
  case fCUR:
    if( mbw_isdigit(c[1]) ) {
      return 1;
    } else if( SET2(c+1) && mbw_isdigit(c[2]) ) {
      cef_state = fCUR;
      return 1;
    }
    cef_state = fDEF;
    return 2;
  case fSEP_1:
    if( c[0] != c[1] ) {
      cef_state = fDEF;
    }
    return 0;
  case fSEP_2:
    cef_state = fSEP_1;
    return 1;
  case fSEP_3:
    cef_state = fSEP_2;
    return 1;
  case fUTF8_1:
    cef_state = fDEF;
    return 2;
  case fUTF8_2:
    cef_state = fUTF8_1;
    return 1;
  case fUTF8_3:
    cef_state = fUTF8_2;
    return 1;
  case fUTF8_4:
    cef_state = fUTF8_3;
    return 1;
  case fUTF8_5:
    cef_state = fUTF8_4;
    return 1;
  case fShift_JIS_1:
  case fBIG5_1:
  case fBIG5P_1:
  case fEUC_CN_1:
  case fJohab_1:
    cef_state = fDEF;
    return 2;
  case fEUC_TW_1:
    cef_state = fDEF;
    return 2;
  case fEUC_TW_2:
    cef_state = fEUC_TW_1;
    return 1;
  case fEUC_TW_3:
    cef_state = fEUC_TW_2;
    return 1;
  case fEUC_Japanese_1:
    cef_state = fDEF;
    return 2;
  case fEUC_Japanese_2:
    cef_state = fEUC_Japanese_1;
    return 1;
  }
  cef_state = fDEF;
  return 0; /* otherwise compiler complains */
}

/* 
 * this code generates good_char() and w_good_char() 
 * returns true if the character is part of a token 
 */
bool_t mbw_prefix(good_char)(mbw_t *c) {
  if( m_options & (1<<M_OPTION_CHAR_ALPHA) ) {
    if( mbw_isalpha(*c) ) {
      if( !(m_options & (1<<M_OPTION_CASEN)) ) {
	*c = mbw_tolower(*c);
      }
      return 1;
    } else {
      return 0;
    }
  } else if( m_options & (1<<M_OPTION_CHAR_CEF) ) {
    return mbw_prefix(is_cef_char)(c);
  } else if( m_options & (1<<M_OPTION_CHAR_ALNUM) ) {
    if( mbw_isalnum(*c) ) {
      if( !(m_options & (1<<M_OPTION_CASEN)) ) {
	*c = mbw_tolower(*c);
      }
      return 1;
    } else {
      return 0;
    }
  } else if( m_options & (1<<M_OPTION_CHAR_GRAPH) ) {
    if( mbw_isgraph(*c) ) {
      if( !(m_options & (1<<M_OPTION_CASEN)) ) {
	*c = mbw_tolower(*c);
      }
      return 1;
    } else {
      return 0;
    }
  }
  return 0;
}

/* 
 * this code generates b64_line_filter2() and w_b64_line_filter2() 
 * works ok so long as q <= line, or q >> line 
 * WARNING: it is assumed that the buffer at q can hold (at most) all of line
 */
mbw_t *mbw_prefix(b64_line_filter2)(mbw_t *line, mbw_t *q) {
  mbw_t *p = line;
  mbw_t buf[4];
  mbw_t *buf_start = buf;
  mbw_t *buf_end = buf + 4;

  if( q ) {
    while( *p ) {
      if( mbw_prefix(b64_code)(*p) > -1 ) {
	*buf_start++ = *p;
	if( buf_start == buf_end ) {
	  buf_start = buf;
	  *q++ = (mbw_prefix(b64_code)(buf[0])<<2) + (mbw_prefix(b64_code)(buf[1])>>4);
	  if( buf[2] != mbw_lit('=') ) {
	    *q++ = (mbw_prefix(b64_code)(buf[1])<<4) + (mbw_prefix(b64_code)(buf[2])>>2);
	    if( buf[3] != mbw_lit('=') ) {
	      *q++ = (mbw_prefix(b64_code)(buf[2])<<6) + mbw_prefix(b64_code)(buf[3]);	    
	    } else {
	      break;
	    }
	  } else {
	    break; 
	  }
	}
      }
      p++;
    }
    *q = mbw_lit('\0');
  }
  return q;
}

/* 
 * this code generates b64_line_filter() and w_b64_line_filter() 
 * Decodes a base64 encoded line. The input line is overwritten. 
 *
 * The b64 standard arbitrarily truncates lines to 57 characters, so
 * here we place the chunks in a cache and only overwrite line when
 * the cache is full. Unfortunately, malformed email messages may not 
 * follow the standard, so in practice all this means is that we get 
 * arbitrarily truncated input.
 *
 * Note that when we overwrite line with the cached data, we assume
 * the line is big enough to hold all the cached data. This is guaranteed
 * by registering the current line length with the cache.
 */
bool_t mbw_prefix(b64_line_filter)(mbw_t *line) {
  mbw_prefix(adjust_cache_size)(&mbw_prefix(b64_dc), mbw_strlen(line));
  mbw_prefix(b64_dc).data_ptr = 
    mbw_prefix(b64_line_filter2)(line, mbw_prefix(b64_dc).data_ptr);
  return mbw_prefix(flush_cache)(&mbw_prefix(b64_dc), line, 0);
}


/* 
 * this code generates qp_line_filter2() and w_qp_line_filter2() 
 * this works ok so long as q <= line, or q >> line 
 * WARNING: it is assumed that the buffer at q can hold (at most) all of line
 */
mbw_t *mbw_prefix(qp_line_filter2)(mbw_t *line, mbw_t *q) {
  mbw_t *p = line;
  if( q ) {
    while( *p ) {
      if( *p != mbw_lit('=') ) {
	*q++ = *p++;
      } else {
	if( !*(++p) || mbw_isspace(*p) ) { 
	  break;
	} else {
 	  /* if the equal sign isn't followed by  */
          /* an upper case hex number, something's wrong */ 
	  *q = mbw_prefix(qp_code)(*p);
	  if( ((signed char)*q < 0) || !p[1] || (mbw_prefix(qp_code)(p[1]) < 0) ) {
	    *q++ = p[-1];
	  } else {
	    *q = (*q << 4) + mbw_prefix(qp_code)(p[1]); 
	    if( *q ) { q++; } 
	    p += 2;
	  }
	}
      }
    }
    *q = mbw_lit('\0');
  }
  return q;
}

/* 
 * this code generates qp_line_filter() and w_qp_line_filter() 
 * Decodes a quoted-printable line. The input line is overwritten. 
 *
 * The QP standard arbitrarily truncates lines to 76 characters, so
 * here we place the chunks in a cache and only overwrite line when
 * the cache is full. Unfortunately, malformed email messages may not 
 * follow the standard, so in practice all this means is that we get 
 * arbitrarily truncated input.
 *
 * Note that when we overwrite line with the cached data, we assume
 * the line is big enough to hold all the cached data. This is guaranteed
 * by registering the current line length with the cache.
 */
bool_t mbw_prefix(qp_line_filter)(mbw_t *line) {
  mbw_prefix(adjust_cache_size)(&mbw_prefix(qp_dc), mbw_strlen(line));
  mbw_prefix(qp_dc).data_ptr = 
    mbw_prefix(qp_line_filter2)(line, mbw_prefix(qp_dc).data_ptr);
  return mbw_prefix(flush_cache)(&mbw_prefix(qp_dc), line, 0);
}

/* 
 * this code generates mhe_line_filter() and w_mhe_line_filter() 
 * translates a MIME message header extension encoded
 * token into its equivalent byte sequence. 
 */
bool_t mbw_prefix(mhe_line_filter)(mbw_t *line) {
  mbw_t *p = line;
  mbw_t *q = line;
  mbw_t *r;

  return 1;
  while( *p ) {
    if( (p[0] == mbw_lit('=')) && (p[1] == mbw_lit('?')) ) {
      r = p + 2;
      while( *r && (*r != mbw_lit('?'))) { r++; }
      r++;
      /* I think lower case is illegal */
      if( (*r == mbw_lit('Q')) || (*r == mbw_lit('q')) ) { 
	if( *(++r) == mbw_lit('?') ) { 
	  r++;
	  /* we are now committed. find end marker and replace with NUL */
	  for(p = r; *p; p++) { 
	    if( *p == mbw_lit('_') ) {
	      *p = ' ';
	    } else if( (p[0] == mbw_lit('?')) && (p[1] == mbw_lit('=')) ) {
	      *p = mbw_lit('\0');
	      break;
	    }
	  }
	  q = mbw_prefix(qp_line_filter2)(r, q);
	  p += 2;
	} else {
	  /* malformed encoding */
	  *q++ = *p++;
	}
      } else if( (*r == mbw_lit('B')) || (*r == mbw_lit('b')) ) { 
	/* I think lower case is illegal, but we're lenient */
	if( *(++r) == mbw_lit('?') ) { 
	  r++;
	  /* we are now committed. find end marker and replace with NUL */
	  for(p = r; *p; p++) { 
	    if( (p[0] == mbw_lit('?')) && (p[1] == mbw_lit('=')) ) {
	      *p = mbw_lit('\0');
	      break;
	    }
	  }
	  q = mbw_prefix(b64_line_filter2)(r, q);
	  p += 2;
	} else {
	  /* malformed encoding */
	  *q++ = *p++;
	}
      } else {
	/* malformed encoding */
	*q++ = *p++;
      }
    } else {
      *q++ = *p++;      
    }
  }
  *q = '\0';
  return 1;
}

/* 
 * this code generates extract_header_label() and w_extract_header_label() 
 * truncates the line after the first colon, thereby filtering out all but
 * the RFC header name.
 */
int mbw_prefix(extract_header_label)(mbw_t *line) {
  mbw_t *p = line;

  if( m_options & (1<<M_OPTION_XHEADERS) ) {
    /* I'm not sure if censoring the Date fields is a good idea */
    if( (mbw_strncmp(p, mbw_lit("X-DBACL"),7) == 0) ||
	(mbw_strncmp(p, mbw_lit("Date:"),5) == 0) ||
	(mbw_strncmp(p, mbw_lit("Path:"),5) == 0) ||
	(mbw_strncmp(p, mbw_lit("Posted:"),7) == 0) ||
	(mbw_strncmp(p, mbw_lit("Expires:"),8) == 0) ||
	(mbw_strncmp(p, mbw_lit("Resent-Date:"),5) == 0) ||
	(mbw_strncmp(p, mbw_lit("Received:"),9) == 0) ||
	(mbw_strncmp(p, mbw_lit("Delivery-Date:"),14) == 0) ||
	(mbw_isspace(line[0]) && mbox.skip_header) ) {
      mbox.skip_header = 1;
      return 0;
    } else {
      mbox.skip_header = 0;
      mbw_prefix(mhe_line_filter)(line);
      return 1;
    }
  }

  return 0;
}

/***********************************************************
 * MBOX PARSING FUNCTIONS                                  *
 ***********************************************************/

/* 
 * this code generates extract_header_label() and w_extract_header_label() 
 * retrieves the MIME boundary if one is found. Doesn't cope with rfc2184 
 */
bool_t mbw_prefix(extract_mime_boundary)(mbw_t *line) {
  const mbw_t *q;
  mbw_t *r;
  int size;
  bool_t quoted = 0; /* used both in calsulation and for return value */

  q = mbw_prefix(mystrcasestr)(line, mbw_lit("boundary="));

  if( q ) {
    /* we skip white space after = sign, even though it is not allowed */
    for(q += 9; mbw_isspace(*q); q++);
    if( *q ) {
      quoted = (*q == mbw_lit('"'));      
      r = mbox.boundary.mbw_prefix(identifier)[mbox.boundary.index];
      size = 0;
      if( quoted ) { 
	for(q++; *q && (*q != mbw_lit('"')) && 
	      (size < MAX_BOUNDARY_BUFSIZE); q++, size++) {
	  *r++ = *q;
	}
      } else {
	for(; *q && !mbw_isspace(*q) && 
	      (size < MAX_BOUNDARY_BUFSIZE); q++, size++) {
	  *r++ = *q;
	}
      }
      mbox.boundary.size[mbox.boundary.index] = size;
      if( ++mbox.boundary.index >= MAX_BOUNDARIES ) { mbox.boundary.index = 0; }
      quoted = (size > 0) ? 1 : 0;
    } else {
      /* this is bad */
      quoted = 0;
    }
  }

  return quoted;
}

/* 
 * this code generates check_mime_boundary() and w_check_mime_boundary() 
 * The check is only approximate.
 */
bool_t mbw_prefix(check_mime_boundary)(const mbw_t *line) {
  int c = (mbox.boundary.index > 0) ? 
    (mbox.boundary.index - 1) : (MAX_BOUNDARIES - 1);
  int k = 0;
  const mbw_t *p = mbox.boundary.mbw_prefix(identifier)[c];
  const mbw_t *q = line + 2; 

  while(*q) {
    if( (k >= mbox.boundary.size[c]) || (*q != *p) ) {
      c--;
      if( c < 0 ) { c = MAX_BOUNDARIES - 1; }
      if( c == mbox.boundary.index ) {
	mbox.boundary.was_end = 0;
	return 0;
      }
      p = &mbox.boundary.mbw_prefix(identifier)[c][k];
    } else if( k == mbox.boundary.size[c] - 1) {
      if((q[1] == mbw_lit('-')) && (q[2] == mbw_lit('-'))) {
	mbox.boundary.was_end = 1;
      } else {
	mbox.boundary.was_end = 0;
      }
      return 1;
    } else {
      /* normally, a space isn't allowed in the boundary, but we're lenient */ 
      q++; 
      p++;
      k++;
    }
  }

  mbox.boundary.was_end = 0;
  return 0;
}

/* 
 * this code generates check_mime_boundary() and w_check_mime_boundary() 
 */
void mbw_prefix(extract_mime_types)(mbw_t *line, MIME_Struct *ms) {
  const mbw_t *q;
  if( !mbw_strncasecmp(line, mbw_lit("Content-Type:"), 13) ) {
    line += 13;
    q = mbw_prefix(mystrcasestr)(line, mbw_lit("text"));
    if( q ) {
      q += 5;
      if( !mbw_strncasecmp(q, mbw_lit("enriched"), 8) ) {
	ms->type = ctRICHTEXT;
      } else if( !mbw_strncasecmp(q, mbw_lit("html"), 4) ) {
	ms->type = ctHTML;
      } else if( !mbw_strncasecmp(q, mbw_lit("plain"), 5) ) {
	ms->type = ctPLAINTEXT;
      } else if( !mbw_strncasecmp(q, mbw_lit("richtext"), 8) ) {
	ms->type = ctRICHTEXT;
      } else if( !mbw_strncasecmp(q, mbw_lit("xml"), 3) ) {
	ms->type = ctXML;
      } else if( !mbw_strncasecmp(q, mbw_lit("sgml"), 4) ) {
	ms->type = ctSGML;
      } else {
	ms->type = ctPLAINTEXT;
      }
      return;
    } 
    
    q = mbw_prefix(mystrcasestr)(line, mbw_lit("multipart"));
    if( q ) {
      ms->type = ctPLAINTEXT;
      return;
    }

    q = mbw_prefix(mystrcasestr)(line, mbw_lit("message"));
    if( q ) {
      q += 8;
      if( !mbw_strncasecmp(q, mbw_lit("delivery-status"), 15) ||
		 !mbw_strncasecmp(q, mbw_lit("http"), 4) ||
		 !mbw_strncasecmp(q, mbw_lit("sip"), 3) ||
		 !mbw_strncasecmp(q, mbw_lit("sipfrag"), 3) ||
		 !mbw_strncasecmp(q, mbw_lit("s-http"), 6) ) {
	ms->type = ctOTHER;
      } else {
	ms->type = ctPLAINTEXT;
      }
      return;
    }

    q = mbw_prefix(mystrcasestr)(line, mbw_lit("application"));
    if( q ) {
      q += 11;
      if( !mbw_strncasecmp(q, mbw_lit("sgml"), 4) ||
	  !mbw_strncasecmp(q, mbw_lit("xml"), 4) ) {
	ms->type = ctPLAINTEXT;
      } else {
	ms->type = ctOTHER;
      }
      return;
    }

    ms->type = ctOTHER;

  } else if( !mbw_strncasecmp(line, mbw_lit("Content-Transfer-Encoding:"), 
			      26) ) {
    line += 26;
    if( mbw_prefix(mystrcasestr)(line, mbw_lit("base64")) ) {
      ms->encoding = ceB64;
    } else if( mbw_prefix(mystrcasestr)(line, mbw_lit("quoted-printable")) ) {
      ms->encoding = ceQP;
    } else if( mbw_prefix(mystrcasestr)(line, mbw_lit("binary")) ) {
      ms->encoding = ceBIN;
    } else {
      ms->encoding = ceID;
    }

  }
}


HEADER_Type mbw_prefix(scan_header_type)(mbw_t *line) {

#define STRIP(q) {while(*q++) { if( *q == mbw_lit(';') ) { *q = mbw_lit('\0'); break; }}}

  if( mbw_isspace(*line) ) {
    if( mbox.strip_header_comment ) { STRIP(line); }
    return htCONT;
  } else if( !mbw_strncasecmp(line, mbw_lit("From:"),5) || 
	     !mbw_strncasecmp(line, mbw_lit("To:"),3) ||
	     !mbw_strncasecmp(line, mbw_lit("Message-ID:"),11) ||
	     !mbw_strncasecmp(line, mbw_lit("In-Reply-To:"),12) ||
	     !mbw_strncasecmp(line, mbw_lit("Subject:"),8) ) {
    STRIP(line);
    mbox.strip_header_comment = 1;
    return htSTANDARD;
  } else if( !mbw_strncasecmp(line, mbw_lit("Content-"),8) &&
	     mbw_strchr(line + 8, mbw_lit(':')) ) {
    mbox.strip_header_comment = 0;
    return htMIME;
  } else if( !mbw_strncasecmp(line, mbw_lit("Sender:"),7) ||
	     !mbw_strncasecmp(line, mbw_lit("Reply-To:"),9) ||
	     !mbw_strncasecmp(line, mbw_lit("Bcc:"),4) ||
	     !mbw_strncasecmp(line, mbw_lit("Cc:"),3) ||
	     !mbw_strncasecmp(line, mbw_lit("References:"),11) ) {
    STRIP(line);
    mbox.strip_header_comment = 1;
    return htEXTENDED;
  } else {
    /* if the line starts with a word missing a :, then 
       it could be a malformed continuation line */
    while( *line && !mbw_isspace(*line) && (*line != mbw_lit(':')) ) { line++; }
    mbox.strip_header_comment = 0;
    return ( *line == mbw_lit(':') ) ? htUNDEF : htCONT;
  }
}

int mbw_prefix(extract_mime_label)(mbw_t *line) {
  if( m_options & (1<<M_OPTION_HEADERS) ) {
    if( !mbw_strncasecmp(line, mbw_lit("Content-"),8) ) {
      line += 8;
      if( !mbw_strncasecmp(line, mbw_lit("Type:"),5) ||
	  !mbw_strncasecmp(line, mbw_lit("Disposition:"),12) ) {
	STRIP(line);
	return 1;
      } else if( !mbw_strncasecmp(line, mbw_lit("Description:"),12) ) {
	/* note: we only get first line of description */
	return 1;
      }
    }
  }
  return 0;
}

/* 
 * this code generates mbox_line_filter() and w_mbox_line_filter() 
 *
 * returns true if the line should be processed further
 * depends on global mbox state 
 */
bool_t mbw_prefix(mbox_line_filter)(mbw_t *line) {
  bool_t line_empty = 0;
  bool_t process_line = 0; /* by default we skip the line */
  XML_Reset force_filter = xmlUNDEF;

  line_empty = ((line[0] == mbw_lit('\n')) ||
		((line[0] == mbw_lit('\r')) && 
		 (line[1] == mbw_lit('\n')))) ? 1 : 0;

  /* STEP 1: first perform state transitions */
  switch(mbox.state) {
  case msUNDEF:
    if( mbox.prev_line_empty && !mbw_strncmp(line, mbw_lit("From "), 5) ) {
      mbox.state = msHEADER;
      mbox.substate = msuUNDEF;
      mbox.skip_until_boundary = 0;
    }
    break;
  case msHEADER:
    if( line_empty ) {
      mbox.state = msBODY;
      mbox.substate = msuUNDEF;
      /* don't reset skip_until_boundary */
      mbox.corruption_check = 5;
    } 
    break; 
  case msBODY:
    if( (line[0] == mbw_lit('-')) && (line[1] == mbw_lit('-')) && 
	!mbw_isspace(line[2]) ) {
      if( mbw_prefix(check_mime_boundary)(line) ) {
	mbox.state = msATTACH;
	mbox.substate = msuUNDEF;
	mbox.skip_until_boundary = mbox.boundary.was_end;
	mbox.corruption_check = 0;
      } else if( mbox.prev_line_empty &&
		 !(m_options & (1<<M_OPTION_PLAIN)) ) {
	/* could be a corrupted boundary - note 
	 * previous empty line is not required, but often true
	 */
	mbox.corruption_check = 5;
      }
    } else if( mbox.prev_line_empty &&
	       !mbw_strncmp(line, mbw_lit("From "), 5) ) {
      mbox.state = msHEADER;
      mbox.substate = msuUNDEF;
      mbox.skip_until_boundary = 0;
      mbox.corruption_check = 0;
    } else if( mbox.corruption_check > 0 ) {
      mbox.corruption_check--;
      /* we filter out mail header extension codings - shouldn't do any harm */
      mbw_prefix(mhe_line_filter)(line); 
      switch(mbw_prefix(scan_header_type)(line)) {
      case htMIME:
	mbox.state = msATTACH;
	mbox.substate = msuMIME;
	mbox.skip_until_boundary = 0;
	mbox.corruption_check = 0;
	break;
      default:
	/* do nothing - so far so good */
	break;
      }
    }
    break;
  case msATTACH:
    if( line_empty ) {
      mbox.state = msBODY;
      mbox.substate = msuUNDEF;
      mbox.skip_until_boundary = 0;
      mbox.corruption_check = 0;
    } 
    break;
  }

  mbox.prev_line_empty = line_empty; /* for next time */

  /* STEP 2: now clean up and prepare the line according to current state
   * and substate. 
   * After cleanup, the variable process_line indicates if the line
   * should be ignored. 
   * The substate can evolve while the current state is unchanging.
   */
  switch(mbox.state) {
  case msUNDEF:
    /* line is not processed */
    break;
  case msHEADER:
    switch(mbox.substate) {
    case msuUNDEF:
      /* flush caches */
      process_line = 
	mbw_prefix(flush_cache)(&mbw_prefix(b64_dc), line, 1) || 
	mbw_prefix(flush_cache)(&mbw_prefix(qp_dc), line, 1);
      if( process_line ) {
	/* we still remember previous type/encoding, decide if we need filter */
	force_filter = select_xml_defaults(&mbox.body);
      }
      /* there are no default mime types for headers */
      mbox.header.type = mbox.body.type = ctUNDEF;
      mbox.header.encoding = mbox.body.encoding = ceUNDEF;
      /* switch to normal state next time */
      mbox.substate = msuOTHER;
      mbox.corruption_check = 0;
      mbox.skip_header = 0;
      mbox.plainstate = psPLAIN;
      break;
    case msuTRACK:
    case msuMIME:
    case msuOTHER:
      /* switch substate if necessary */
      switch(mbw_prefix(scan_header_type)(line)) {
      case htSTANDARD:
	mbox.substate = msuTRACK;
	break;
      case htEXTENDED:
	mbox.substate = (m_options & (1<<M_OPTION_HEADERS)) ? msuTRACK : msuOTHER;
	break;
      case htMIME:
	mbox.substate = msuMIME;
	break;
      case htCONT:
	/* nothing */
	break;	
      case htUNDEF:
	mbox.substate = msuOTHER;
	break;
      }

      /* process substate */
      
      switch(mbox.substate) {
      case msuTRACK:
	process_line = mbw_prefix(mhe_line_filter)(line);
	break;
      case msuMIME:
	mbw_prefix(mhe_line_filter)(line);
	mbw_prefix(extract_mime_types)(line, &mbox.header);
	mbox.skip_until_boundary = 
	  mbw_prefix(extract_mime_boundary)(line) || mbox.skip_until_boundary;
	/* this comes last, modifies line */
	process_line = 0; /* mbw_prefix(extract_mime_label)(line); */
	break;
      case msuUNDEF:
      case msuOTHER:
	process_line = mbw_prefix(extract_header_label)(line);
	break;
      }
    }
    break;
  case msBODY:
    switch(mbox.substate) {
    case msuUNDEF:
      /* flush caches */
      process_line = 
	mbw_prefix(flush_cache)(&mbw_prefix(b64_dc), line, 1) || 
	mbw_prefix(flush_cache)(&mbw_prefix(qp_dc), line, 1);
      if( process_line ) {
	/* we still remember previous type/encoding, decide if we need filter */
	force_filter = select_xml_defaults(&mbox.body);
      }
      /* bodies by default inherit the header mime types */
      if( mbox.body.type == ctUNDEF ) 
	{ mbox.body.type = mbox.header.type; }
      if( mbox.body.encoding == ceUNDEF ) 
	{ mbox.body.encoding = mbox.header.encoding; }

      /* switch to normal state next time */
      mbox.substate = msuTRACK;
      mbox.plainstate = psPLAIN;
      break;
    case msuTRACK:
    case msuMIME:
    case msuOTHER:
      if( mbox.skip_until_boundary ) {
	process_line = 0;
      } else {
	switch(mbox.body.type) {
	case ctUNDEF:
	  /* the header didn't say, so we must assume text */
	case ctPLAINTEXT:
	case ctRICHTEXT:
	case ctHTML:
	case ctXML:
	case ctSGML:
	case ctUNTEXT:
	  switch(mbox.body.encoding) {
	  case ceBIN:
	    process_line = 1;
	    break;
	  case ceUNDEF:
	  case ceID:
	    process_line = ((m_options & (1<<M_OPTION_PLAIN)) ?
			    1 : mbw_prefix(plain_text_filter)(line));
	    break;
	  case ceQP:
 	    process_line = mbw_prefix(qp_line_filter)(line);
	    break;
	  case ceB64:
	    process_line = mbw_prefix(b64_line_filter)(line);
	    break;
	  }
	  break;
	case ctOTHER:
	  process_line = 0;
	  break;
	}
      }
      break;
    }
    break;
  case msATTACH:
    switch(mbox.substate) {
    case msuUNDEF:
      /* flush caches */
      process_line = 
	mbw_prefix(flush_cache)(&mbw_prefix(b64_dc), line, 1) || 
	mbw_prefix(flush_cache)(&mbw_prefix(qp_dc), line, 1);
      if( process_line ) {
	/* we still remember previous type/encoding, decide if we need filter */
	force_filter = select_xml_defaults(&mbox.body);
      }
      /* attachments by default inherit the header mime types */
      mbox.body.type = mbox.header.type;
      mbox.body.encoding = mbox.header.encoding;
      /* switch to normal state next time */
      mbox.substate = msuTRACK;
      mbox.plainstate = psPLAIN;
      break;
    case msuTRACK:
    case msuMIME:
    case msuOTHER:
      /* switch substate if necessary */
      switch(mbw_prefix(scan_header_type)(line)) {
      case htSTANDARD:
	mbox.substate = msuTRACK;
	break;
      case htEXTENDED:
	mbox.substate = (m_options & (1<<M_OPTION_HEADERS)) ? msuTRACK : msuOTHER;
	break;
      case htMIME:
	mbox.substate = msuMIME;
	break;
      case htCONT:
	/* nothing */
	break;	
      case htUNDEF:
	mbox.substate = msuOTHER;
	break;
      }
      /* process substate */
      switch(mbox.substate) {
      case msuTRACK:
      case msuUNDEF:
      case msuOTHER:
	process_line = 0;
	break;
      case msuMIME:
	mbw_prefix(mhe_line_filter)(line);
	mbw_prefix(extract_mime_types)(line, &mbox.body);
	mbox.skip_until_boundary = 
	  mbw_prefix(extract_mime_boundary)(line) || mbox.skip_until_boundary;
	/* this comes last, modifies line */
	process_line = 0; /* mbw_prefix(extract_mime_label)(line); */
	break;
      }
      break;
    }
    break;
  }

  /* STEP 3: activate filters */
  if( force_filter != xmlUNDEF ) {
    reset_xml_character_filter(force_filter);
  } else {
    if( mbox.state == msBODY ) {
      if( mbox.skip_until_boundary ) {
	process_line = 0;
      } else {
	reset_xml_character_filter(select_xml_defaults(&mbox.body));
      }
    } else {
      reset_xml_character_filter(xmlDISABLE);
    }
  }
  return process_line;
}


/***********************************************************
 * HTML PARSING FUNCTIONS                                  *
 ***********************************************************/

/* 
 * this code generates check_mime_boundary() and w_check_mime_boundary() 
 * (this is ugly, but I am _not_ building a string hash, sheesh).
 *
 * note: the conversion from unicode to multibyte depends on the current
 * locale, but also assumes that wchar_t *is* unicode internally. Both assumptions
 * can be false on weird compilers. In case the locale is incapable of doing the job, 
 * we convert based on the hex code.
 */
void mbw_prefix(decode_html_entity)(mbw_t **lline, mbw_t **qq) {
  mbw_t *line = *lline;
  mbw_t *q = *qq;

#if defined HAVE_MBRTOWC

  mbw_t *r = NULL;
#if defined MBW_MB
  int s,t;
  mbw_t scratch[16]; /* C compiler complains  about MB_CUR_MAX */
#endif
  wchar_t c = 0; /* this must always be wchar_t */

  switch(line[1]) {
  case mbw_lit('#'):
    if( (line[2] == mbw_lit('x')) || (line[2] == mbw_lit('X')) ) {
      c = (wchar_t)mbw_strtol(line + 3, &r, 16);
    } else {
      c = (wchar_t)mbw_strtol(line + 2, &r, 10);
    }
    break;

#define ENTITY(x,y,z) if( !mbw_strncmp((line + 3), (x + 2), (y - 2)) ) \
                           { c = (z); r = line + (y) + 1; }

  case mbw_lit('a'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("aacute"),6,L'\xe1');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("acute"),5,L'\xb4') else ENTITY(mbw_lit("acirc"),4,L'\xe2');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("aelig"),5,L'\xe6');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("agrave"),6,L'\xe0');
      break;
    case mbw_lit('l'):
      ENTITY(mbw_lit("alpha"),5,L'\x3b1') else ENTITY(mbw_lit("alefsym"),7,L'\x2135');
      break;
    case mbw_lit('m'):
      ENTITY(mbw_lit("amp"),3,L'\x26');
      break;
    case mbw_lit('n'):
      ENTITY(mbw_lit("ang"),3,L'\x2220') else ENTITY(mbw_lit("and"),3,L'\x2227');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("aring"),5,L'\xe5');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("asymp"),5,L'\x2248');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("atilde"),6,L'\xe3');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("auml"),4,L'\xe4');
      break;
    }
    break;

  case mbw_lit('A'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Aacute"),6,L'\xc1');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("Acirc"),5,L'\xc2');
      break;
    case mbw_lit('E'):
      ENTITY(mbw_lit("AElig"),5,L'\xc6');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("Agrave"),6,L'\xc0');
      break;
    case mbw_lit('l'):
      ENTITY(mbw_lit("Alpha"),5,L'\x391');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("Aring"),5,L'\xc5');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("Atilde"),6,L'\xc3');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("Auml"),4,L'\xc4');
      break;
    }
    break;

  case mbw_lit('b'):
    switch(line[2]) {
    case mbw_lit('d'):
      ENTITY(mbw_lit("bdquo"),5,L'\x201e');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("beta"),4,L'\x3b2');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("bull"),4,L'\x2022');
      break;
    }
    break;

  case mbw_lit('B'):
    switch(line[2]) {
    case mbw_lit('e'):
      ENTITY(mbw_lit("Beta"),4,L'\x392');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("Brvbar"),6,L'\xa6');
      break;
    }
    break;

  case mbw_lit('c'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("cap"),3,L'\x2229');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("ccedil"),6,L'\xe7');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("cent"),4,L'\xa2') else ENTITY(mbw_lit("cedil"),5,L'\xb8');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("chi"),3,L'\x3c7');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("circ"),4,L'\x2c6');
      break;
    case mbw_lit('l'):
      ENTITY(mbw_lit("clubs"),5,L'\x2663');
      break;
    case mbw_lit('o'):
      ENTITY(mbw_lit("copy"),4,L'\xa9') else ENTITY(mbw_lit("cong"),4,L'\x2245');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("crarr"),5,L'\x21b5');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("curren"),6,L'\xa4') else ENTITY(mbw_lit("cup"),3,L'\x222a');
      break;
    }
    break;

  case mbw_lit('C'):
    switch(line[2]) {
    case mbw_lit('c'):
      ENTITY(mbw_lit("Ccedil"),6,L'\xc7');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("Chi"),3,L'\x3a7');
      break;
    }
    break;

  case mbw_lit('d'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("darr"),4,L'\x2193') else ENTITY(mbw_lit("dagger"),6,L'\x2020');
      break;
    case mbw_lit('A'):
      ENTITY(mbw_lit("dArr"),4,L'\x21d3');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("delta"),5,L'\x3b4');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("divide"),6,L'\xf7') else ENTITY(mbw_lit("diams"),5,L'\x2666');
      break;
    }
    break;

  case mbw_lit('D'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Dagger"),6,L'\x2021');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("Deg"),3,L'\xb0') else ENTITY(mbw_lit("Delta"),5,L'\x394');
      break;
    }
    break;

  case mbw_lit('e'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("eacute"),6,L'\xe9');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("ecirc"),5,L'\xea');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("egrave"),6,L'\xe8');
      break;
    case mbw_lit('m'):
      ENTITY(mbw_lit("empty"),5,L'\x2205') else ENTITY(mbw_lit("emsp"),4,L'\x2003');
      break;
    case mbw_lit('n'):
      ENTITY(mbw_lit("ensp"),4,L'\x2002');
      break;
    case mbw_lit('p'):
      ENTITY(mbw_lit("epsilon"),7,L'\x3b5');
      break;
    case mbw_lit('q'):
      ENTITY(mbw_lit("equiv"),5,L'\x2261');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("eth"),3,L'\xf0') else ENTITY(mbw_lit("eta"),3,L'\x3b7');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("euml"),4,L'\xeb') else ENTITY(mbw_lit("euro"),4,L'\x20ac');
      break;
    case mbw_lit('x'):
      ENTITY(mbw_lit("exist"),5,L'\x2203');
      break;
    }
    break;

  case mbw_lit('E'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Eacute"),6,L'\xc9');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("Ecirc"),5,L'\xca');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("Egrave"),6,L'\xc8');
      break;
    case mbw_lit('p'):
      ENTITY(mbw_lit("Epsilon"),7,L'\x395');
      break;
    case mbw_lit('T'):
      ENTITY(mbw_lit("ETH"),3,L'\xd0');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("Eta"),3,L'\x397');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("Euml"),4,L'\xcb');
      break;
    }
    break;

  case mbw_lit('f'):
    switch(line[2]) {
    case mbw_lit('n'):
      ENTITY(mbw_lit("fnof"),4,L'\x192');
      break;
    case mbw_lit('o'):
      ENTITY(mbw_lit("forall"),6,L'\x2200');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("frac14"),6,L'\xbc') else ENTITY(mbw_lit("frac12"),6,L'\xbd') else
	ENTITY(mbw_lit("frac34"),6,L'\xbe') else ENTITY(mbw_lit("frasl"),5,L'\x2044');
      break;
    }
    break;

  case mbw_lit('F'):
    /* nothing */
    break;

  case mbw_lit('g'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("gamma"),5,L'\x3b3');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("ge"),2,L'\x2265');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("gt"),2,L'\x3e');
      break;
    }
    break;

  case mbw_lit('G'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Gamma"),5,L'\x393');
      break;
    }
    break;

  case mbw_lit('h'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("harr"),4,L'\x2194');
      break;
    case mbw_lit('A'):
      ENTITY(mbw_lit("hArr"),4,L'\x21d4');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("hearts"),6,L'\x2665') else ENTITY(mbw_lit("hellip"),6,L'\x2026');
      break;
    }
    break;

  case mbw_lit('H'):
    /* nothing */
    break;

  case mbw_lit('i'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("iacute"),6,L'\xed');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("icirc"),5,L'\xee');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("iexcl"),5,L'\xa1');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("igrave"),6,L'\xec');
      break;
    case mbw_lit('m'):
      ENTITY(mbw_lit("image"),5,L'\x2111');
      break;
    case mbw_lit('n'):
      ENTITY(mbw_lit("infin"),5,L'\x221e') else ENTITY(mbw_lit("int"),3,L'\x222b');
      break;
    case mbw_lit('o'):
      ENTITY(mbw_lit("iota"),4,L'\x3b9');
      break;
    case mbw_lit('q'):
      ENTITY(mbw_lit("iquest"),6,L'\xbf');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("isin"),4,L'\x2208');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("iuml"),4,L'\xef');
      break;
    }
    break;

  case mbw_lit('I'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Iacute"),6,L'\xcd');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("Icirc"),5,L'\xce');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("Igrave"),6,L'\xcc');
      break;
    case mbw_lit('o'):
      ENTITY(mbw_lit("Iota"),4,L'\x399');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("Iuml"),4,L'\xcf');
      break;
    }
    break;

  case mbw_lit('j'):
    /* nothing */
    break;

  case mbw_lit('J'):
    /* nothing */
    break;

  case mbw_lit('k'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("kappa"),5,L'\x3ba');
      break;
    }
    break;

  case mbw_lit('K'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Kappa"),5,L'\x39a');
      break;
    }
    break;

  case mbw_lit('l'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("lambda"),6,L'\x3bb') else ENTITY(mbw_lit("lang"),4,L'\x2329');
      break;
    case mbw_lit('A'):
      ENTITY(mbw_lit("lArr"),4,L'\x21d0');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("lceil"),5,L'\x2308');
      break;
    case mbw_lit('d'):
      ENTITY(mbw_lit("ldquo"),5,L'\x201c');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("le"),2,L'\x2264');
      break;
    case mbw_lit('f'):
      ENTITY(mbw_lit("lfloor"),6,L'\x2309');
      break;
    case mbw_lit('o'):
      ENTITY(mbw_lit("lowast"),6,L'\x2217') else ENTITY(mbw_lit("loz"),3,L'\x25ca');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("lrm"),3,L'\x200e');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("lsquo"),5,L'\x2018') else ENTITY(mbw_lit("lsaquo"),6,L'\x2039');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("lt"),2,L'\x3c');
      break;
    }
    break;

  case mbw_lit('L'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Laquo"),5,L'\xab') else ENTITY(mbw_lit("Lambda"),6,L'\x39b') else
	ENTITY(mbw_lit("Larr"),4,L'\x2190');
      break;
    }
    break;

  case mbw_lit('m'):
    switch(line[2]) {
    case mbw_lit('d'):
      ENTITY(mbw_lit("mdash"),5,L'\x2014');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("minus"),5,L'\x2212');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("mu"),2,L'\x3bc');
      break;
    }
    break;

  case mbw_lit('M'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Macr"),4,L'\xaf');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("Micro"),5,L'\xb5') else ENTITY(mbw_lit("Middot"),6,L'\xb7');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("Mu"),2,L'\x39c');
      break;
    }
    break;

  case mbw_lit('n'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("nabla"),5,L'\x2207');
      break;
    case mbw_lit('b'):
      ENTITY(mbw_lit("nbsp"),4,L'\xa0');
      break;
    case mbw_lit('d'):
      ENTITY(mbw_lit("ndash"),5,L'\x2013');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("ne"),2,L'\x2260');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("ni"),2,L'\x220b');
      break;
    case mbw_lit('o'):
      ENTITY(mbw_lit("not"),3,L'\xac') else ENTITY(mbw_lit("notin"),5,L'\x2209');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("nsub"),4,L'\x2284');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("ntilde"),6,L'\xf1');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("nu"),2,L'\x3bd');
      break;
    }
    break;

  case mbw_lit('N'):
    switch(line[2]) {
    case mbw_lit('t'):
      ENTITY(mbw_lit("Ntilde"),6,L'\xd1');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("Nu"),2,L'\x39d');
      break;
    }
    break;

  case mbw_lit('o'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("oacute"),6,L'\xf3');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("ocirc"),5,L'\xf4');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("oelig"),5,L'\x153');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("ograve"),6,L'\xf2');
      break;
    case mbw_lit('l'):
      ENTITY(mbw_lit("oline"),5,L'\x203e');
      break;
    case mbw_lit('m'):
      ENTITY(mbw_lit("omicron"),7,L'\x3bf') else ENTITY(mbw_lit("omega"),5,L'\x3c9');
      break;
    case mbw_lit('p'):
      ENTITY(mbw_lit("oplus"),5,L'\x2295');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("ordf"),4,L'\xaa') else ENTITY(mbw_lit("ordm"),4,L'\xba') else
	ENTITY(mbw_lit("or"),2,L'\x2228');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("oslash"),6,L'\xf8');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("otilde"),6,L'\xf5') else ENTITY(mbw_lit("otimes"),6,L'\x2297');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("ouml"),4,L'\xf6');
      break;
    }
    break;

  case mbw_lit('O'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Oacute"),6,L'\xd3');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("Ocirc"),5,L'\xd4');
      break;
    case mbw_lit('E'):
      ENTITY(mbw_lit("OElig"),5,L'\x152');
      break;
    case mbw_lit('m'):
      ENTITY(mbw_lit("Omicron"),7,L'\x39f') else ENTITY(mbw_lit("Omega"),5,L'\x3a9');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("Ograve"),6,L'\xd2');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("Oslash"),6,L'\xd8');
      break;
    case mbw_lit('t'):
      ENTITY(mbw_lit("Otilde"),6,L'\xd5');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("Ouml"),4,L'\xd6');
      break;
    } 
    break;

  case mbw_lit('p'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("part"),4,L'\x2202');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("perp"),4,L'\x22a5') else ENTITY(mbw_lit("permil"),6,L'\x2030');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("phi"),3,L'\x3c6');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("pi"),2,L'\x3c0') else ENTITY(mbw_lit("piv"),3,L'\x3d6');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("prime"),5,L'\x2032') else	ENTITY(mbw_lit("prod"),4,L'\x220f');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("psi"),3,L'\x3c8');
      break;
    }
    break;

  case mbw_lit('P'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Para"),4,L'\xb6');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("Phi"),3,L'\x3a6');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("Pi"),2,L'\x3a0');
      break;
    case mbw_lit('l'):
      ENTITY(mbw_lit("Plusmn"),6,L'\xb1');
      break;
    case mbw_lit('o'):
      ENTITY(mbw_lit("Pound"),5,L'\xa3');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("Prime"),5,L'\x2033') else	ENTITY(mbw_lit("Prop"),4,L'\x221d');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("Psi"),3,L'\x3a8');
      break;
    }
    break;

  case mbw_lit('q'):
    switch(line[2]) {
    case mbw_lit('u'):
      ENTITY(mbw_lit("quot"),4,L'\x22');
      break;
    }
    break;

  case mbw_lit('Q'):
    /* nothing */
    break;

  case mbw_lit('r'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("rarr"),4,L'\x2192') else ENTITY(mbw_lit("radic"),5,L'\x221a') else
	ENTITY(mbw_lit("rang"),4,L'\x232a');
      break;
    case mbw_lit('A'):
      ENTITY(mbw_lit("rArr"),4,L'\x21d2');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("rceil"),5,L'\x2309');
      break;
    case mbw_lit('d'):
      ENTITY(mbw_lit("rdquo"),5,L'\x201d');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("real"),4,L'\x211C') else ENTITY(mbw_lit("reg"),3,L'\xae');
      break;
    case mbw_lit('f'):
      ENTITY(mbw_lit("rfloor"),6,L'\x230a');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("rho"),3,L'\x3c1');
      break;
    case mbw_lit('l'):
      ENTITY(mbw_lit("rlm"),3,L'\x200f');
      break;
    case mbw_lit('s'):
      ENTITY(mbw_lit("rsquo"),5,L'\x2019') else	ENTITY(mbw_lit("rsaquo"),6,L'\x203a');
      break;
    }
    break;

  case mbw_lit('R'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Raquo"),5,L'\xbb');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("Reg"),3,L'\xae');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("Rho"),3,L'\x3a1');
      break;
    }
    break;

  case mbw_lit('s'):
    switch(line[2]) {
    case mbw_lit('b'):
      ENTITY(mbw_lit("sbquo"),5,L'\x201a');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("scaron"),6,L'\x161');
      break;
    case mbw_lit('d'):
      ENTITY(mbw_lit("sdot"),4,L'\x22c5');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("sect"),4,L'\xa7');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("shy"),3,L'\xad');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("sigmaf"),6,L'\x3c2') else	ENTITY(mbw_lit("sigma"),5,L'\x3c3') else
	ENTITY(mbw_lit("sim"),3,L'\x223c');
      break;
    case mbw_lit('p'):
      ENTITY(mbw_lit("spades"),6,L'\x2660');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("sup2"),4,L'\xb2') else ENTITY(mbw_lit("sup3"),4,L'\xb3') else
	ENTITY(mbw_lit("sup1"),4,L'\xb9') else ENTITY(mbw_lit("sum"),3,L'\x2211') else
	  ENTITY(mbw_lit("sub"),3,L'\x2282') else ENTITY(mbw_lit("sup"),3,L'\x2283') else
	    ENTITY(mbw_lit("sube"),4,L'\x2286') else ENTITY(mbw_lit("supe"),4,L'\x2287');
      break;
    case mbw_lit('z'):
      ENTITY(mbw_lit("szlig"),5,L'\xdf');
      break;
    }
    break;

  case mbw_lit('S'):
    switch(line[2]) {
    case mbw_lit('c'):
      ENTITY(mbw_lit("Scaron"),6,L'\x160');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("Sigma"),5,L'\x3a3');
      break;
    }
    break;

  case mbw_lit('t'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("tau"),3,L'\x3c4');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("thorn"),5,L'\xfe') else ENTITY(mbw_lit("theta"),5,L'\x3b8') else
	ENTITY(mbw_lit("thetasym"),8,L'\x3d1') else ENTITY(mbw_lit("there4"),6,L'\x2234') else
	  ENTITY(mbw_lit("thinsp"),6,L'\x2009');
      break;
    case mbw_lit('i'):
      ENTITY(mbw_lit("times"),5,L'\xd7') else ENTITY(mbw_lit("tilde"),5,L'\x2dc');
      break;
    case mbw_lit('r'):
      ENTITY(mbw_lit("trade"),5,L'\x2122');
      break;
    }
    break;

  case mbw_lit('T'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Tau"),3,L'\x3a4');
      break;
    case mbw_lit('h'):
      ENTITY(mbw_lit("Theta"),5,L'\x398');
      break;
    case mbw_lit('H'):
      ENTITY(mbw_lit("THORN"),5,L'\xde');
      break;
    }
    break;

  case mbw_lit('u'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("uacute"),6,L'\xfa') else ENTITY(mbw_lit("uarr"),4,L'\x2191');
      break;
    case mbw_lit('A'):
      ENTITY(mbw_lit("uArr"),4,L'\x21d1');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("ucirc"),5,L'\xfb');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("ugrave"),6,L'\xf9');
      break;
    case mbw_lit('m'):
      ENTITY(mbw_lit("uml"),3,L'\xa8');
      break;
    case mbw_lit('p'):
      ENTITY(mbw_lit("upsilon"),7,L'\xc5') else	ENTITY(mbw_lit("upsih"),5,L'\x3d2');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("uuml"),4,L'\xfc');
      break;
    }
    break;

  case mbw_lit('U'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Uacute"),6,L'\xda');
      break;
    case mbw_lit('c'):
      ENTITY(mbw_lit("Ucirc"),5,L'\xdb');
      break;
    case mbw_lit('g'):
      ENTITY(mbw_lit("Ugrave"),6,L'\xd9');
      break;
    case mbw_lit('p'):
      ENTITY(mbw_lit("Upsilon"),7,L'\xa5');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("Uuml"),4,L'\xdc');
      break;
    }
    break;

  case mbw_lit('v'):
    /* nothing */
    break;

  case mbw_lit('V'):
    /* nothing */
    break;

  case mbw_lit('w'):
    switch(line[2]) {
    case mbw_lit('e'):
      ENTITY(mbw_lit("weierp"),6,L'\x2118');
      break;
    }
    break;

  case mbw_lit('W'):
    /* nothing */
    break;

  case mbw_lit('x'):
    switch(line[2]) {
    case mbw_lit('i'):
      ENTITY(mbw_lit("xi"),2,L'\x3be');
      break;
    }
    break;

  case mbw_lit('X'):
    switch(line[2]) {
    case mbw_lit('i'):
      ENTITY(mbw_lit("Xi"),2,L'\x39e');
      break;
    }
    break;

  case mbw_lit('y'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("yacute"),6,L'\xfd');
      break;
    case mbw_lit('e'):
      ENTITY(mbw_lit("yen"),3,L'\xa5');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("yuml"),4,L'\xff');
      break;
    }
    break;

  case mbw_lit('Y'):
    switch(line[2]) {
    case mbw_lit('a'):
      ENTITY(mbw_lit("Yacute"),6,L'\xdd');
      break;
    case mbw_lit('u'):
      ENTITY(mbw_lit("Yuml"),4,L'\x178');
      break;
    }
    break;

  case mbw_lit('z'):
    switch(line[2]) {
    case mbw_lit('e'):
      ENTITY(mbw_lit("zeta"),4,L'\x3b6');
      break;
    case mbw_lit('w'):
      ENTITY(mbw_lit("zwnj"),4,L'\x200c') else ENTITY(mbw_lit("zwj"),3,L'\x200d');
      break;
    }
    break;

  case mbw_lit('Z'):
    switch(line[2]) {
    case mbw_lit('e'):
      ENTITY(mbw_lit("Zeta"),4,L'\x396');
      break;
    }
    break;

  default:
    break;
  }

#if defined MBW_WIDE
  if( c && r ) {
    *q++ = c;
    line = r;
  } else {
    *q++ = *line;
  }
#else

  /* now if c is nonzero, then we found the entity */
  if( c && r ) {
    if( c == L'\xa0' ) { 
      /* shortcut for &nbsp; */
      *q++ = mbw_lit(' ');
    } else {
      s = wcrtomb(scratch,c,NULL);
      if( (s > -1) && (q + s <= r) ) {
	for(t = 0; t < s; t++) { *q++ = scratch[t]; }
      } else {
	/* locale doesn't recognize this char */
	s = c - L'\0';
	if( s < 0xFF ) { *q++ = (mbw_t)s; }
      }
    }
    line = r;
  } else {
    *q++ = *line; /* copy just a single char */
  }
#endif

#else /* HAVE_MBRTOWC is not defined */

  /* we don't have wchar_t, just copy the token */
  *q++ = *line;

#endif

  *lline = line;
  *qq = q;
}

/* 
 * this code generates decode_escaped_uri_character() and 
 * w_decode_escaped_uri_character().
 */
void mbw_prefix(decode_uri_character)(mbw_t **lline, mbw_t **qq) {
  mbw_t *line = *lline;
  mbw_t *q = *qq;
  mbw_t scratch[3];
  mbw_t c = 0;
  mbw_t *r;

  if( *line == mbw_lit('%') ) {
    /* check that the next two chars are hex */
    for(r = line + 1; mbw_isspace(*r); r++);
    if( mbw_isxdigit(*r) ) {
      scratch[0] = *r;
      for(r++; mbw_isspace(*r); r++);
      if( mbw_isxdigit(*r) ) {
	scratch[1] = *r;
	scratch[2] = mbw_lit('\0');
	c = (mbw_t)mbw_strtol(line + 1, NULL, 16);
      }
    }

    if( c ) {
      *q++ = c;
      line = r;
    } else {
      *q++ = *line;
    }

  } else if( mbw_isspace(*line) ) {
    /* ignore */
  } else {
    /* not an escaped character */
    *q++ = *line;
  }

  *lline = line;
  *qq = q;
}

/* this reads one or more characters from line and outputs 
 * zero or one character at q. 
 * The characters output depend on the xml.attribute state, but
 * the function never outputs more than it reads. Thus if q <= line
 * to begin with, this is preserved. This is necessary as the function
 * is normally used to modify line in-place.
 *
 * For URL type attributes, this function assumes the standard URI
 * form scheme:netloc/extra, and only prints the netloc part.
 */
void mbw_prefix(xml_attribute_filter)(mbw_t **lline, mbw_t **qq) {
  switch(xml.attribute) {
  case SRC:
    /* we don't output anything until the first : */
    if( *(*lline) == mbw_lit(':') ) { 
      xml.attribute = SRC_NETLOC_PREFIX; 
    }
    break;
  case SRC_NETLOC_PREFIX:
    switch(*(*lline)) {
    case mbw_lit('/'):
      /* ignore */
      break;
    case mbw_lit('?'):
    case mbw_lit(';'):
    case mbw_lit('#'):
    case mbw_lit('&'):
      xml.attribute = SRC_NETLOC_SUFFIX;
      break;
    default:
      mbw_prefix(decode_uri_character)(lline,qq);
      xml.attribute = SRC_NETLOC;
      break;
    }
    break;
  case SRC_NETLOC:
    switch(*(*lline)) {
    case mbw_lit('/'):
      *(*qq)++ = *(*lline);
      xml.attribute = SRC_NETLOC_PATH;
      break;
    case mbw_lit('?'):
    case mbw_lit(';'):
    case mbw_lit('#'):
    case mbw_lit('&'):
      xml.attribute = SRC_NETLOC_SUFFIX;
      break;
    default:
      mbw_prefix(decode_uri_character)(lline,qq);
      break;
    }
    break;
  case SRC_NETLOC_PATH:
    switch(*(*lline)) {
    case mbw_lit('.'):
      *(*qq)++ = mbw_lit(' ');
      break;
    case mbw_lit('?'):
    case mbw_lit(';'):
    case mbw_lit('#'):
    case mbw_lit('&'):
      xml.attribute = SRC_NETLOC_SUFFIX;
      break;
    default:
      mbw_prefix(decode_uri_character)(lline,qq);
      break;
    }
    break;
  case SRC_NETLOC_SUFFIX:
    /* nothing */
    break;
  case ALT:
    if( *(*lline) == mbw_lit('&') ) {
      mbw_prefix(decode_html_entity)(lline, qq);
    } else {
      *(*qq)++ = *(*lline);
    }
    break;
  case UNDEF:
    /* nothing */
    break;
  }
}

/* removes tags in the string - modifies in place 
 * the name of this function is a misnomer, since it doesn't
 * parse xml properly. But we just want a simple kludge 
 * for most html flavours 
 */
void mbw_prefix(xml_character_filter)(mbw_t *line) {
  mbw_t *q;
  q = line;

  /* don't call this with y < 1 */
#define TAGMATCH(x,y) (!mbw_strncasecmp(line + 1, x + 1, y - 1) && (mbw_isspace(line[y]) || (line[y] == mbw_lit('>')) || (line[y] == mbw_lit('\0'))) && (line += (y - 1)))

  /* this is convenient for debugging */
#define PDEBUG(x) printf(#x"{%c%c%c%c%c%c%c%c%c%c}\n", line[0], line[1], line[2], line[3], line[4], line[5], line[6], line[7], line[8], line[9])

  while( *line ) {
    switch(xml.state) {
    case TEXT:
      switch(line[0]) {
      case mbw_lit('<'):
	/* does it look like <x where x is either alpha or punctuation? */
	if( mbw_isalpha(line[1]) ) {
	  line++;
	  /* tags aren't mined, xtags are */
	  xml.state = TAG; 
	  xml.attribute = UNDEF;

	  switch(mbw_tolower(line[0])) {
	  case mbw_lit('a'):
	    if( mbw_isspace(line[1]) || 
		TAGMATCH(mbw_lit("area"),4) ||
		TAGMATCH(mbw_lit("applet"),6) ) {
	      xml.state = XTAG;
	    }
	    break;
	  case mbw_lit('b'):
	    if( TAGMATCH(mbw_lit("base"),4) ||
		TAGMATCH(mbw_lit("bgsound"),7) ) {
	      xml.state = XTAG;
	    } else if( TAGMATCH(mbw_lit("br"),2) ) {
	      *q++ = mbw_lit(' ');
	      xml.state = TAG;
	    } else if( TAGMATCH(mbw_lit("body"),4) ) {
	      xml.hide = VISIBLE;
	      if( xml.parser == xpSMART ) {
		xml.parser = xpHTML;
	      }
	    }
	    break;
	  case mbw_lit('c'):
	    if( TAGMATCH(mbw_lit("comment"),7) ) {
	      if( xml.parser == xpHTML ) {
		xml.hide = COMMENT;
	      }
	    }
	    break;
	  case mbw_lit('d'):
	    if( TAGMATCH(mbw_lit("div"),3) ) {
	      xml.state = XTAG;
	    }
	    break;
	  case mbw_lit('e'):
	    if( TAGMATCH(mbw_lit("embed"),5) ) {
	      xml.state = XTAG;
	    }
	    break;
	  case mbw_lit('f'):
	    if( TAGMATCH(mbw_lit("frame"),5) ) {
	      xml.state = XTAG;
	    }
	    break;
	  case mbw_lit('h'):
	    if( TAGMATCH(mbw_lit("html"),4) || TAGMATCH(mbw_lit("head"),4) ) {
	      xml.hide = VISIBLE;
	      if( xml.parser == xpSMART ) {
		xml.parser = xpHTML;
	      }
	    } else if( TAGMATCH(mbw_lit("hr"),2) ) {
	      *q++ = mbw_lit(' ');
	      xml.state = TAG;
	    }
	    break;
	  case mbw_lit('i'):
	    if( TAGMATCH(mbw_lit("img"),3) ||
		TAGMATCH(mbw_lit("iframe"),6) ||
		TAGMATCH(mbw_lit("ilayer"),6) ||
		TAGMATCH(mbw_lit("input"),5) ) {
	      xml.state = XTAG;
	    }
	    break;
	  case mbw_lit('l'):
	    if( TAGMATCH(mbw_lit("layer"),5) ||
		TAGMATCH(mbw_lit("link"),4) ) {
	      xml.state = XTAG;
	    }
	    break;
	  case mbw_lit('n'):
	    if( (TAGMATCH(mbw_lit("noframes"),8) && (xml.hide = NOFRAMES)) ||
		(TAGMATCH(mbw_lit("nolayer"),7) && (xml.hide = NOLAYER)) ||
		(TAGMATCH(mbw_lit("noscript"),8) && (xml.hide = NOSCRIPT)) ||
		(TAGMATCH(mbw_lit("noembed"),7) && (xml.hide = NOEMBED)) ) {
	      if( xml.parser == xpHTML ) {
		if( (m_options & (1<<M_OPTION_SHOW_ALT)) ) {		  
		  xml.hide = VISIBLE; 
		}
	      }
	    }
	    break;
	  case mbw_lit('s'):
	    if( TAGMATCH(mbw_lit("span"),4) ) {
	      xml.state = XTAG;
	    } else if( TAGMATCH(mbw_lit("script"),6) ) {
	      xml.hide = SCRIPT;
	    } else if( TAGMATCH(mbw_lit("style"),5) ) {
	      xml.hide = STYLE;
	    }
	    break;
	  case mbw_lit('t'):
	    if( TAGMATCH(mbw_lit("title"),5) ) {
		xml.hide = TITLE;
	    }
	    break;
	  default:
	    /* ignore, ie it's a TAG */
	    break;
	  }
	} else if( line[1] == mbw_lit('/') ) {
	  line++;
	  /* tags aren't mined, xtags are */
	  xml.state = TAG; 
	  xml.attribute = UNDEF;

	  switch(mbw_tolower(line[1])) {
	  case mbw_lit('b'):
	    if( TAGMATCH(mbw_lit("/body"),5) ) {
	      if( xml.parser == xpHTML ) {
		xml.parser = xpSMART; 
	      }
	      xml.hide = VISIBLE;
	    }
	    break;
	  case mbw_lit('c'):
	    if( (xml.hide == COMMENT) && TAGMATCH(mbw_lit("/comment"),8) ) {
	      /* nothing */
	    }
	    break;
	  case mbw_lit('h'):
	    if( TAGMATCH(mbw_lit("/html"),5) || TAGMATCH(mbw_lit("/head"),5) ) {
	      if( xml.parser == xpHTML ) {
		xml.parser = xpSMART; 
	      }
	      xml.hide = VISIBLE;
	    }
	    break;
	  case mbw_lit('n'):
	    if( ((xml.hide == NOFRAMES) && TAGMATCH(mbw_lit("/noframes"),9)) ||
		((xml.hide == NOSCRIPT) && TAGMATCH(mbw_lit("/noscript"),9)) ||
		((xml.hide == NOLAYER) && TAGMATCH(mbw_lit("/nolayer"),8)) ||
		((xml.hide == NOEMBED) && TAGMATCH(mbw_lit("/noembed"),8)) ) {
	      xml.hide = VISIBLE;
	    }
	    break;
	  case mbw_lit('s'):
	    if( ((xml.hide == SCRIPT) && TAGMATCH(mbw_lit("/script"),7)) ||
		((xml.hide == STYLE) && TAGMATCH(mbw_lit("/style"),6)) ) {
	      xml.hide = VISIBLE;
	    }
	    break;
	  case mbw_lit('t'):
	    if( TAGMATCH(mbw_lit("/title"),6) ) {
	      xml.hide = VISIBLE;
	    }
	    break;
	  default:
	    /* ignore, ie it's a TAG */
	    break;
	  }
	} else { /* second char is not alpha or slash */
	  if( mbw_strncmp(line + 1, mbw_lit("!--"), 3) == 0 ) {
	    xml.state = CMNT;
	    line += 3;
	  } else if( line[1] == mbw_lit('<') ) {
	    /* stay in TEXT state */
	    if( xml.parser == xpDUMB ) {
	      line++;
	    } else {
	      if( (xml.hide == VISIBLE) || 
		  (xml.hide == TITLE) ||
		  ((m_options & (1<<M_OPTION_SHOW_SCRIPT)) && (xml.hide == SCRIPT)) ) {
		*q++ = *line;
	      }
	    }
	  } else {
	    /* bogus tag? */
	    xml.state = TAG;
	    line++;
	  }
	}
	break;
      case mbw_lit('&'):
	if( (xml.parser == xpHTML) || (xml.parser == xpDUMB) ) {
	  if( (xml.hide == VISIBLE) || 
	      (xml.hide == TITLE) ||
	      ((m_options & (1<<M_OPTION_SHOW_SCRIPT)) && (xml.hide == SCRIPT)) ) {
	    mbw_prefix(decode_html_entity)(&line, &q); 
	  }
	}
	break;
      default:
	if( (xml.parser == xpHTML) || (xml.parser == xpDUMB) ) {
	  if( (xml.hide == VISIBLE) || 
	      (xml.hide == TITLE) ||
	      ((m_options & (1<<M_OPTION_SHOW_SCRIPT)) && (xml.hide == SCRIPT)) ) {
	    *q++ = *line; 
	  }
	}
	break;
      }
      break;
    case TAG:
      if( line[0] == mbw_lit('>') ) {
	xml.state = TEXT;
      } else if( line[0] == mbw_lit('\'') ) {
	xml.state = TAGQUOTE;
      } else if( line[0] == mbw_lit('"') ) {
	xml.state = TAGDQUOTE;
      }
      break;
    case TAGQUOTE:
      if( (line[0] == '\\') && (line[1] == mbw_lit('\'')) ) {
	line++;
      } else if( line[0] == mbw_lit('\'') ) {
	xml.state = TAG;
      }
      break;
    case TAGDQUOTE:
      if( (line[0] == '\\') && (line[1] == mbw_lit('"')) ) {
	line++;
      } else if( line[0] == mbw_lit('"') ) {
	xml.state = TAG;
      }
      break;
    case XTAG:
      if( xml.parser == xpSMART ) { 
	/* we've recognized an HTML tag */
	xml.parser = xpHTML; 
      }
      if( (xml.attribute != UNDEF) && mbw_isspace(line[0]) ) {
	xml.attribute = UNDEF;
	*q++ = mbw_lit(']');
      } else
      switch(mbw_tolower(line[0])) {
      case mbw_lit('>'):
	xml.state = TEXT;
	if( xml.attribute != UNDEF ) {
	  xml.attribute = UNDEF;
	  *q++ = mbw_lit(']');
	}
	break;
      case mbw_lit('\''):
	xml.state = XTAGQUOTE;
	break;
      case mbw_lit('"'):
	xml.state = XTAGDQUOTE;
	break;
      case mbw_lit('a'):
	mbw_prefix(xml_attribute_filter)(&line,&q);
	if( m_options & (1<<M_OPTION_SHOW_ALT) ) {
	  if( !mbw_strncasecmp(line, mbw_lit("alt="), 4) ) {
	    if( xml.attribute != UNDEF ) {
	      *q++ = mbw_lit(']');
	    }
	    xml.attribute = ALT;
	    line += 3;
	    *q++ = mbw_lit('[');
	  } 
	}
	break;
      case mbw_lit('h'):
	mbw_prefix(xml_attribute_filter)(&line,&q);
	if( m_options & (1<<M_OPTION_SHOW_LINKS) ) {
	  if( !mbw_strncasecmp(line, mbw_lit("href="), 5) ) {
	    if( xml.attribute != UNDEF ) {
	      *q++ = mbw_lit(']');
	    }
	    xml.attribute = SRC;
	    line += 4;
	    *q++ = mbw_lit('[');
	  } 
	}
	break;
      case mbw_lit('s'):
	mbw_prefix(xml_attribute_filter)(&line,&q);
	if( m_options & (1<<M_OPTION_SHOW_LINKS) ) {
	  if( !mbw_strncasecmp(line, mbw_lit("src="), 4) ) {
	    if( xml.attribute != UNDEF ) {
	      *q++ = mbw_lit(']');
	    }
	    xml.attribute = SRC;
	    line += 3;
	    *q++ = mbw_lit('[');
	  } 
	}
	break;
      case mbw_lit('t'):
	mbw_prefix(xml_attribute_filter)(&line,&q);
	if( m_options & (1<<M_OPTION_SHOW_ALT) ) {
	  if( !mbw_strncasecmp(line, mbw_lit("title="), 6) ) {
	    if( xml.attribute != UNDEF ) {
	      *q++ = mbw_lit(']');
	    }
	    xml.attribute = ALT;
	    line += 5;
	    *q++ = mbw_lit('[');
	  } 
	}
	break;
      default:
	mbw_prefix(xml_attribute_filter)(&line,&q);
	break;
      }
      break;
    case XTAGQUOTE:
      if( line[0] == mbw_lit('\'') ) {
	xml.state = XTAG;
      } else {
	mbw_prefix(xml_attribute_filter)(&line,&q);
	if( (line[0] == '\\') && (line[1] == mbw_lit('\'')) ) {
	  line++;
	  mbw_prefix(xml_attribute_filter)(&line,&q);
	}
      }
      break;
    case XTAGDQUOTE:
      if( line[0] == mbw_lit('"') ) {
	xml.state = XTAG;
      } else {
	mbw_prefix(xml_attribute_filter)(&line,&q);
	if( (line[0] == '\\') && (line[1] == mbw_lit('"')) ) {
	  line++;
	  mbw_prefix(xml_attribute_filter)(&line,&q);
	}
      }
      break;
    case CMNT:
      if( (line[0] == mbw_lit('-')) && 
	  (!mbw_strncmp(line + 1, mbw_lit("->"), 2) && (line += 2)) ) {
	xml.state = TEXT;
      } else {
	/* ignore comments in some circumstances */
	if( (m_options & (1<<M_OPTION_SHOW_HTML_COMMENTS)) ||
	    ((m_options & (1<<M_OPTION_SHOW_SCRIPT)) && (xml.hide = SCRIPT)) ) {
	  *q++ = *line;
	}
      }
      break;
    case DISABLED:
      /* don't modify the line at all */
      return;
    }
    line++;
  }
  *q = '\0'; /* mark the end of the clean text string */
}


/***********************************************************
 * TEXT PARSING FUNCTIONS                                  *
 ***********************************************************/

int mbw_prefix(is_uuline)(mbw_t *line) {
  int i = 0;
  int len = (int)(line[0] - mbw_lit(' '));

  if( (len < 0) || (len > 63) ) {
    return -1;
  }
  while(len-- > 0) {
    if( (line[i] > mbw_lit('`')) || 
	(line[i] < mbw_lit(' ')) ) {
      return -2;
    }
    i++;
  }
  return 1;
}

bool_t mbw_prefix(plain_text_filter)(mbw_t *line) {
  mbw_t *q;
  bool_t url = 0;
  bool_t censor = 0;

#define ISOCT(x) ((x <= mbw_lit('6')) && (x >= mbw_lit('0')))

  switch(mbox.plainstate) {
  case psPLAIN:
    if( (line[0] == mbw_lit('b')) && 
	!mbw_strncmp(line, mbw_lit("begin "),6) && 
	ISOCT(line[6]) && ISOCT(line[7]) && ISOCT(line[8]) ) {
      mbox.plainstate = psUUENCODE;
      return 1;
    }
    break;
  case psUUENCODE:
    switch(mbw_prefix(is_uuline)(line)) {
    case -1:
      return 1;
    case -2:
      mbox.plainstate = psPLAIN;
      break;
    default:
      return 0;
    }
    break;
  }

  /* now assume psPLAIN */

  q = line;
  while(*line) {
    switch(*line) {
    case mbw_lit('%'):
      if( !censor ) {
	mbw_prefix(decode_uri_character)(&line, &q);
      }
      break;
    case mbw_lit('&'):
      if( !censor ) {
	if( !url ) {
	  mbw_prefix(decode_html_entity)(&line, &q);
	} else {
	  censor = 1;
	}
      }
      break;
    case mbw_lit('H'):
    case mbw_lit('h'):
      if( !mbw_strncasecmp(line, mbw_lit("http://"), 7) ) {
	censor = 0;
	url = 1;
      }
      if( !censor ) { *q++ = *line; }
      break;
    case mbw_lit('?'):
    case mbw_lit(';'):
    case mbw_lit('#'):
      if( url ) { 
	censor = 1; 
      } else {
	*q++ = *line;
      }
      break;
    case mbw_lit(' '):
    case mbw_lit('\t'):
    case mbw_lit('>'):
    case mbw_lit('\''):
    case mbw_lit('"'):
      if( censor && url ) { censor = 0; url = 0; }
      *q++ = *line;
      break;
    default:
      if( !censor ) { *q++ = *line; }
      break;
    }
    line++;
  }
  *q = '\0';
  return 1;
}

#endif


